/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.datasources.interlib.service

import groovy.util.logging.Slf4j;
import metridoc.datasources.interlib.util.DateUtil;

@Slf4j
class BorrowDirectService {
	public static String BD_SERVICE_KEY = 'bd'
	public static String EZB_SERVICE_KEY = 'ezb'

	static ConfigObject config
	static{
		GroovyClassLoader classLoader = new GroovyClassLoader(BorrowDirectService.class.getClassLoader())
		try {
			config = new ConfigSlurper().parse(classLoader.loadClass('BorrowDirectQueries'));
		} catch (Exception e) {
			println "Failed to load queries"
			e.printStackTrace();
		}
	}

	private static getSummaryDates(fiscalYear, result){
		def currentFiscalYear = fiscalYear;
		Date currentFiscalYearEnd;
		Date lastFiscalYearEnd;
		if(currentFiscalYear == null){
			def currentDate = Calendar.getInstance();
			def currentYear = currentDate.get(Calendar.YEAR);
			def currentMonth = currentDate.get(Calendar.MONTH);
			currentFiscalYear = DateUtil.getFiscalYear(currentYear, currentMonth)
			if(result)
				result.currentMonth = currentMonth

			lastFiscalYearEnd = DateUtil.getDate(currentYear-1, currentMonth,currentDate.get(Calendar.DAY_OF_MONTH), 0, 0, 0);
			currentFiscalYearEnd = DateUtil.getDate(currentYear, currentMonth,currentDate.get(Calendar.DAY_OF_MONTH), 0, 0, 0);
		}else{
			if(result)
				result.currentMonth = DateUtil.getFiscalYearEndMonth();

			currentFiscalYearEnd = DateUtil.getFiscalYearEndDate(currentFiscalYear)
			lastFiscalYearEnd =  DateUtil.getFiscalYearEndDate(currentFiscalYear - 1)
		}
		if(result)
			result.fiscalYear = currentFiscalYear

		Date currentFiscalYearStart = DateUtil.getFiscalYearStartDate(currentFiscalYear)

		def dates = [:]
		dates.currentFiscalYear = [
			currentFiscalYearStart,
			currentFiscalYearEnd
		];

		dates.lastFiscalYear = [
			DateUtil.getFiscalYearStartDate(currentFiscalYear - 1),
			lastFiscalYearEnd
		]
		return dates;
	}

	public static loadDataPerLibrary(sql, fiscalYear, isBorrowing, result, serviceKey){
		def dates = getSummaryDates(fiscalYear, result)
		loadDataPerLibrary(sql, isBorrowing, result, dates, null, serviceKey, null);
	}

	private static loadDataPerLibrary(sql, isBorrowing, result, dates, libId, tablePrefix, selectedLibIds){
		def libRoleColumn;
		def keyForSection;
		def additionalCondition = ""
		if(isBorrowing){
			if(libId != null){
				additionalCondition = "and " + config.borrowdirect.db.column.borrower+"="+libId
				libRoleColumn = config.borrowdirect.db.column.lender
			}else{
				libRoleColumn = config.borrowdirect.db.column.borrower
			}
			keyForSection = 'borrowing'
		}else{
			if(libId != null){
				additionalCondition = "and " + config.borrowdirect.db.column.lender+"="+libId
				libRoleColumn = config.borrowdirect.db.column.borrower
			}else{
				libRoleColumn = config.borrowdirect.db.column.lender
			}
			keyForSection = 'lending'
		}
		if(selectedLibIds != null){
			additionalCondition += " and " + libRoleColumn + getInClause(selectedLibIds);
		}

		def allQuery = getAdjustedQuery(config.queries.borrowdirect.countsPerLibrary, libRoleColumn, additionalCondition, tablePrefix)
		def query = getAdjustedQuery(config.queries.borrowdirect.countsPerLibraryMonthlyFilled, libRoleColumn, additionalCondition, tablePrefix)

		def allLibDataSection = getLibDataMap(-1l, result).get(keyForSection)
		//currentFiscalYear
		def sqlParams = dates.currentFiscalYear
		log.debug("Runnig query for currentFiscalYear: " + query + " params="+sqlParams)
		sql.eachRow(query, sqlParams, {
			def libData = getLibDataMap(it.getAt(0), result)
			int currentKey = it.getAt(1) != null ? it.getAt(1) : -1
			libData.get(keyForSection).currentFiscalYear.put(currentKey, it.requestsNum)
			if(it.getAt(0) > 0){
				//count sum for each month (row All Libraries)
				def prevValue = allLibDataSection.currentFiscalYear.get(currentKey);
				if( prevValue == null){
					prevValue = 0;
				}
				allLibDataSection.currentFiscalYear.put(currentKey, it.requestsNum + prevValue)
				if(currentKey == -1 && !isBorrowing && libId != null){
					//if there is lending fillRate and for current library it.getAt(0) year number is > 0
					//set default yearFillRate=1, if there were any unfilled requests
					//this number will be updated later (section '//lending: yearFillRate for lib')
					libData.get(keyForSection).yearFillRate = 1
				}
			}
		})
		//turnaround
		def turnaroundQuery = getAdjustedQuery(config.queries.borrowdirect.turnaroundPerLibrary, libRoleColumn, additionalCondition, tablePrefix)
		log.debug("Runnig query for turnaround: " + turnaroundQuery + " params="+sqlParams)
		sql.eachRow(turnaroundQuery, sqlParams, {
			def libData = getLibDataMap(it.getAt(0).longValue(), result)
			def currentMap = libData.get(keyForSection)
			currentMap.turnaroundReqRec = it.turnaroundReqRec
			currentMap.turnaroundReqShp = it.turnaroundReqShp
			currentMap.turnaroundShpRec = it.turnaroundShpRec
		})

		if(isBorrowing){
			if(libId == null){//No lender info for unfilled items
				//borrowing:yearFillRate
				log.debug("Runnig query for yearFillRate: " + allQuery + " params="+sqlParams)
				sql.eachRow(allQuery,
						sqlParams, {
							def libData = getLibDataMap(it.getAt(0), result)
							def currentMap = libData.get(keyForSection)
							if(currentMap.currentFiscalYear.get(-1) == null){
								currentMap.currentFiscalYear.put(-1, 0);
							}
							currentMap.yearFillRate = (it.requestsNum != 0?
									currentMap.currentFiscalYear.get(-1)/(float)it.requestsNum :-1)
						})
			}
		}else if(libId != null){
			//lending: yearFillRate for lib
			def unfilledReqsQuery = getAdjustedQuery(config.queries.borrowdirect.countsPerLibraryUnfilled, libRoleColumn, additionalCondition, tablePrefix)
			def unfilledReqsParams = [
				dates.currentFiscalYear[0],
				dates.currentFiscalYear[1],
				libId]
			log.debug("Runnig query for yearFillRate: lending: " + unfilledReqsQuery + " params="+unfilledReqsParams)
			sql.eachRow(unfilledReqsQuery,
					unfilledReqsParams, {
						def libData = getLibDataMap(it.getAt(0), result)
						def currentMap = libData.get(keyForSection)
						if(currentMap.currentFiscalYear.get(-1) == null){
							currentMap.currentFiscalYear.put(-1, 0);
						}
						def allRequests = currentMap.currentFiscalYear.get(-1) + it.requestsNum
						currentMap.yearFillRate = (allRequests != 0?
								currentMap.currentFiscalYear.get(-1)/(float)allRequests :-1)
					})

		}
		//lastFiscalYear
		sqlParams = dates.lastFiscalYear
		log.debug("Runnig query for lastFiscalYear: " + query+ " params="+sqlParams)
		sql.eachRow(query, sqlParams, {
			def libData = getLibDataMap(it.getAt(0), result)
			int currentKey = it.getAt(1) != null ? it.getAt(1) : -1
			libData.get(keyForSection).lastFiscalYear.put(currentKey, it.requestsNum)
			if(it.getAt(0) > -1){
				def prevValue = allLibDataSection.lastFiscalYear.get(currentKey);
				if( prevValue == null){
					prevValue = 0;
				}
				allLibDataSection.lastFiscalYear.put(currentKey, it.requestsNum + prevValue)
			}
		})
		log.debug("Done for " + keyForSection)
		return result
	}


	public static getLibraryList(sql, serviceKey){
		return getLibraryList(sql, serviceKey, null);
	}
	private static getInClause(paramList){
		return " IN (" + paramList.join(',') + ")"
	}
	public static getLibraryList(sql, serviceKey, selectedLibIds){
		def additionalCondition = selectedLibIds != null?" where library_id " + getInClause(selectedLibIds):""
		def query = prepareQuery(config.queries.borrowdirect.libraryList, serviceKey)
		query = query.replaceAll("\\{add_condition\\}", additionalCondition);
		return sql.rows(query, [])
	}

	private static String getAdjustedQuery(query, libRoleColumn, additionalCondition, tablePrefix){
		def result = query.replaceAll("\\{lib_role\\}", libRoleColumn)
		result = result.replaceAll("\\{add_condition\\}", additionalCondition)
		return prepareQuery(result, tablePrefix);
	}

	private static String prepareQuery(query, tablePrefix){
		return query.replaceAll("\\{table_prefix\\}", tablePrefix)
	}

	private static getLibDataMap(libId, container){
		if(container.get(libId) == null){
			container.put(libId, ['borrowing':['currentFiscalYear':[:], 'lastFiscalYear':[:]],
						'lending':['currentFiscalYear':[:], 'lastFiscalYear':[:]]])
		}
		return container.get(libId)
	}
}